(* $Id: Div.v,v 1.3 2011-10-22 11:09:45 deraugla Exp $ *)

Require Import Utf8.
Require Import Arith.

Require Import gcd_p.

Fixpoint f_mod (a b : nat) : nat :=
  match a with
  | 0 => 0
  | S a' =>
      let r := f_mod a' b in
      if lt_dec (S r) b then S r
      else 0
  end.

Fixpoint f_div (a b : nat) : nat * nat :=
  match a with
  | 0 => (0, 0)
  | S a' =>
      let (q, r) := f_div a' b in
      if lt_dec (S r) b then (q, S r)
      else (S q, 0)
  end.

(* equivalence mod and f_mod *)

Theorem f_mod__mod_p : ∀ a b r, r = f_mod a b → mod_p a b r.
Proof.
fix F 1; intros a b r H2.
destruct a; simpl in H2.
 subst r; apply m_b.

 remember (f_mod a b) as r'.
 apply F in Heqr'.
  remember (lt_dec (S r') b) as H3.
  destruct H3 as [H3| H3]; subst.
   apply m_r; [ apply H3 | apply Heqr' ].

   destruct b.
    apply m_a.

    apply m_q.
    clear HeqH3; apply not_lt in H3.
    remember Heqr' as H4.
    clear HeqH4.
    apply mod_lt in Heqr'.
    apply le_antisym in H3.
     apply eq_add_S in H3.
     subst r'; exact H4.

     destruct Heqr' as [H2| H2].
      exact H2.
      subst; apply lt_0_Sn.
Qed.

Theorem mod_p__f_mod : ∀ a b r, mod_p a b r → r = f_mod a b.
Proof.
induction a; intros b r H1.
 destruct r; [ reflexivity | inversion H1 ].
 simpl.
 inversion H1 as [| | a' b' r' H H2 H3 H4| a' r' H2 H3 H4]; subst.
  reflexivity.

  apply IHa in H2.
  rewrite <- H2.
  remember (lt_dec (S r') b) as H3.
  destruct H3 as [H3| H3].
   reflexivity.
   contradiction H.

  apply IHa in H2.
  rewrite <- H2.
  remember (lt_dec (S r') (S r')) as H3.
  destruct H3 as [H3| H3].
   clear HeqH3; apply lt_irrefl in H3; inversion H3.
   reflexivity.
Qed.

Lemma f_div__eucl : ∀ a b q r,
  0 < b → f_div a b = (q, r) → a = b * q + r ∧ r < b.
Proof.
induction a; intros b q r Hb H1.
 simpl in H1.
 inversion H1; subst; clear H1.
 split.
  rewrite mult_0_r; reflexivity.

  exact Hb.

 simpl in H1.
 remember (f_div a b) as d.
 destruct d as (q₁, r₁).
 remember (lt_dec (S r₁) b) as e.
 destruct e as [e| e].
  inversion H1; subst; clear H1.
  symmetry in Heqd.
  apply IHa in Heqd.
   destruct Heqd as (H1, H2).
   split.
    rewrite <- plus_n_Sm; apply eq_S, H1.
    exact e.
   exact Hb.

  inversion H1; subst; clear H1.
  symmetry in Heqd.
  apply IHa in Heqd.
   destruct Heqd as (H1, H2).
   clear Heqe; apply not_lt in e; unfold ge in e.
   apply le_antisym in e.
    split.
     rewrite plus_0_r.
     subst b.
     simpl.
     rewrite mult_comm.
     simpl.
     subst a.
     simpl.
     rewrite plus_assoc.
     rewrite plus_comm.
     rewrite plus_assoc.
     auto with arith.

     exact Hb.
    exact H2.
   exact Hb.
Qed.

Infix "mod" := f_mod (at level 40).

Lemma f_mod_is_zero : ∀ a b, 0 < b → f_mod a b = 0 → ∃ k, a = k * b.
Proof.
intros a b bpos H.
symmetry  in H.
apply f_mod__mod_p, mod__eucl_div in H.
 destruct H.
 destruct H as (q, H).
 exists q.
 rewrite mult_comm,  <- plus_0_r.
 assumption.
 
 assumption.
Qed.
