(* $Id: Gcd.v,v 1.3 2011-12-01 10:03:59 deraugla Exp $ *)

Require Import Utf8.
Require Import Arith.

Require Import gcd_p.
Require Import Div.

Fixpoint f_ngcd (n a b : nat) : nat :=
  match n with
  | 0 => 0
  | S n' =>
      match b with
      | 0 => a
      | S b' => f_ngcd n' b (f_mod a b)
      end
  end.

Definition gcd (a b : nat) : nat := f_ngcd (S b) a b.

Fixpoint f_nbezout (n a b : nat) : bool * (nat * nat) :=
  match n with
  | 0 => (false, (0, 0))
  | S n' =>
      match b with
      | 0 => (false, (1, 0))
      | S b' =>
          let (q, r) := f_div a b in
          let (is_neg, uv) := f_nbezout n' b r in
          let (u, v) := uv in
          (negb is_neg, (v, u + q * v))
      end
  end.

Definition f_bezout (a b : nat) : bool * (nat * nat) := f_nbezout (S b) a b.

(* equivalence gcd_p and f_gcd *)

Lemma f_ngcd__gcd_p_b_lt_n : ∀ a b n, b < n → f_ngcd n a b = 0 → gcd_p a b 0.
Proof.
fix F 3.
intros a b n H1 H2.
destruct n.
 inversion H1.

 destruct b.
  simpl in H2; subst; apply gcd_O.

  simpl in H2.
  remember (f_mod a (S b)) as r.
  remember Heqr as H3.
  clear HeqH3.
  apply f_mod__mod_p in Heqr.
  apply mod_lt in Heqr.
  apply F in H2.
   destruct Heqr as [H4| H4].
    apply gcd_Sn with (r := f_mod a (S b)).
     apply lt_0_Sn.

     apply f_mod__mod_p; reflexivity.

     rewrite <- H3.
     apply H2.

    subst.
    apply gcd_Sn with (r := 0).
     apply lt_0_Sn.

     apply f_mod__mod_p, H3.

     apply H2.

   destruct Heqr as [H4| H4].
    apply lt_le_trans with (m := S b).
     apply H4.

     apply le_S_n in H1.
     apply H1.

    subst r.
    apply lt_S_n in H1.
    destruct n.
     inversion H1.

     apply lt_0_Sn.
Qed.

Lemma f_ngcd__gcd_p_0 : ∀ a b, f_ngcd (S b) a b = 0 → gcd_p a b 0.
Proof.
intros a b H1.
apply f_ngcd__gcd_p_b_lt_n with (n := S b).
 apply le_n.
 apply H1.
Qed.

Lemma f_ngcd__gcd_p : ∀ a b g n, 0 < g → f_ngcd n a b = g → gcd_p a b g.
Proof.
fix F 4.
intros a b g n H1 H2.
destruct n.
 simpl in H2; subst g; apply lt_irrefl in H1; destruct H1.

 destruct b.
  simpl in H2; subst g; apply gcd_O.

  simpl in H2.
  remember (f_mod a (S b)) as r.
  apply gcd_Sn with (r := r).
   apply lt_0_Sn.
   apply f_mod__mod_p; subst r; reflexivity.
   apply F with (n := n); [ apply H1 | apply H2 ].
Qed.

Theorem f_gcd__gcd_p : ∀ a b g, g = gcd a b → gcd_p a b g.
Proof.
intros a b g H1.
symmetry in H1.
destruct g.
 apply f_ngcd__gcd_p_0, H1.
 apply f_ngcd__gcd_p with (n := S b).
  apply lt_0_Sn.
  apply H1.
Qed.

(* *)

Lemma nbezout_gcd : ∀ n a b, ∀ is_neg u v,
  b < n →
  f_nbezout n a b = (is_neg, (u, v)) →
    (is_neg = false ∧ a * u = b * v + gcd a b) \/
    (is_neg = true ∧ b * v = a * u + gcd a b).
Proof.
fix F 1; intros n a b is_neg u v Hle H.
destruct n.
 apply lt_n_O in Hle; destruct Hle.

 simpl in H.
 destruct b.
  inversion H; subst; clear H.
  left; split; [ reflexivity | idtac ].
  remember (gcd a 0) as g.
  apply f_gcd__gcd_p in Heqg.
  inversion Heqg; subst; [ apply mult_1_r | inversion H ].

  remember (f_div a (S b)) as d.
  destruct d as (q, r).
  symmetry in Heqd; apply f_div__eucl in Heqd.
   destruct Heqd as (H3, H4).
   assert (gcd (S b) r = gcd a (S b)) as Hsame_gcd.
    remember (gcd (S b) r) as g₁.
    remember (gcd a (S b)) as g₂.
    apply f_gcd__gcd_p in Heqg₁.
    apply f_gcd__gcd_p in Heqg₂.
    apply gcd_Sn with (a := a) in Heqg₁.
     apply gcd_uniq with (g1 := g₁) in Heqg₂; [ exact Heqg₂ | exact Heqg₁ ].
     apply lt_0_Sn.
     apply eucl_div__mod; split; [ exists q; exact H3 | exact H4 ].
    rewrite <- Hsame_gcd; clear Hsame_gcd.
    remember (f_nbezout n (S b) r) as bz.
    destruct bz as (is_neg₁, u₁v₁).
    destruct u₁v₁ as (u₁, v₁).
    destruct is_neg₁.
     inversion H; subst; clear H.
     left; split; [ reflexivity | idtac ].
     symmetry in Heqbz; apply F in Heqbz.
      destruct Heqbz as [H| H].
       destruct H as (H1, H2); discriminate H1.

       destruct H as (H1, H2); clear H1.
       simpl.
       rewrite mult_plus_distr_r.
       rewrite H2.
       simpl.
       remember (gcd (S b) r) as g.
       rewrite mult_plus_distr_r.
       rewrite mult_plus_distr_l.
       repeat rewrite plus_assoc.
       rewrite mult_assoc.
       (* omega *)
       clear.
       rewrite plus_comm; symmetry; rewrite plus_comm.
       induction g; [ simpl | apply eq_S, IHg ].
       rewrite <- plus_assoc, plus_comm, <- plus_assoc.
       symmetry.
       rewrite plus_comm.
       induction (b * u₁); [ simpl | apply eq_S, IHn ].
       rewrite <- plus_assoc, plus_comm, <- plus_assoc.
       reflexivity.

      apply le_S_n in Hle.
      apply le_trans with (n := S r) in Hle; [ exact Hle | exact H4 ].

     inversion H; subst; clear H.
     right; split; [ reflexivity | idtac ].
     symmetry in Heqbz; apply F in Heqbz.
      destruct Heqbz as [H| H].
       destruct H as (H1, H2); clear H1.
       simpl.
       rewrite mult_plus_distr_r.
       remember (gcd (S b) r) as g.
       simpl in H2.
       rewrite mult_plus_distr_r.
       rewrite mult_plus_distr_l.
       repeat rewrite plus_assoc.
       rewrite mult_assoc.
       (* omega *)
       symmetry; rewrite <- plus_assoc, <- H2.
       clear.
       rewrite <- plus_assoc.
       symmetry.
       rewrite <- plus_assoc, <- plus_assoc, plus_comm.
       induction (q * u); simpl; [ idtac | apply eq_S, IHn ].
       rewrite <- plus_assoc, plus_comm, <- plus_assoc.
       reflexivity.

       destruct H as (H1, H2); discriminate H1.

      apply le_S_n in Hle.
      apply le_trans with (n := S r) in Hle; [ exact Hle | exact H4 ].

   apply lt_0_Sn.
Qed.

Theorem bezout_gcd : ∀ a b, ∀ is_neg u v,
  f_bezout a b = (is_neg, (u, v)) →
    (is_neg = false ∧ a * u = b * v + gcd a b) \/
    (is_neg = true ∧ b * v = a * u + gcd a b).
Proof.
intros a b is_neg u v H1.
destruct is_neg.
 right.
 split; [ reflexivity | idtac ].
 unfold f_bezout in H1.
 apply nbezout_gcd in H1.
  destruct H1 as [H1| H1].
   destruct H1 as (H1, H2); discriminate H1.
   destruct H1 as (H1, H2); exact H2.
  apply le_n.
 left.
 split; [ reflexivity | idtac ].
 unfold f_bezout in H1.
 apply nbezout_gcd in H1.
  destruct H1 as [H1| H1].
   destruct H1 as (H1, H2); exact H2.
   destruct H1 as (H1, H2); discriminate H1.
  apply le_n.
Qed.

(* *)

Theorem gcd_p__f_gcd : ∀ a b g, gcd_p a b g → g = gcd a b.
Proof.
intros a b g H.
remember (gcd a b) as g₁.
apply f_gcd__gcd_p in Heqg₁.
apply gcd_uniq with (g1 := g₁) in H.
 subst; reflexivity.
 exact Heqg₁.
Qed.

(* *)

Lemma mod_kb_lt : ∀ a b k x, 0 < k → f_mod a (k * b) < x → f_mod a b < x.
Proof.
intros a b k x kpos H.
destruct b.
 rewrite mult_0_r in H; exact H.

 remember (f_mod a (k * S b)) as mk.
 apply f_mod__mod_p in Heqmk.
 apply mod__eucl_div in Heqmk.
  destruct Heqmk as (eucl, lim).
  destruct eucl as (q, eucl).
  remember (f_mod a (S b)) as m.
  apply f_mod__mod_p in Heqm.
  rewrite eucl in Heqm.
  rewrite plus_comm, mult_comm, mult_assoc in Heqm.
  apply mod_plus_kb__mod in Heqm.
  apply mod__eucl_div in Heqm.
   destruct Heqm as (eucl', lim').
   destruct eucl' as (q', eucl').
   apply le_lt_trans with (m := mk).
    subst mk; apply le_plus_r.
    exact H.
   apply lt_0_Sn.
  rewrite mult_comm; simpl.
  apply lt_le_trans with (m := k).
   exact kpos.
   apply le_plus_l.
Qed.

Lemma plus_mod_lt : ∀ a b c,
  a + f_mod b c < c → a + f_mod b c = f_mod (a + b) c.
Proof.
intros a b c apmltc.
remember (f_mod b c) as r₁.
apply f_mod__mod_p in Heqr₁.
apply mod__eucl_div in Heqr₁.
 destruct Heqr₁ as (eucl₁, lim₁).
 destruct eucl₁ as (q₁, eucl₁).
 remember (f_mod (a + b) c) as r₂.
 apply f_mod__mod_p in Heqr₂.
 rewrite eucl₁ in Heqr₂.
 rewrite plus_assoc, plus_comm, plus_assoc, mult_comm in Heqr₂.
 apply mod_plus_kb__mod in Heqr₂.
 apply mod_p_by_gt in Heqr₂.
  subst r₂; apply plus_comm.
  rewrite plus_comm; exact apmltc.
 destruct c.
  apply lt_n_O in apmltc; destruct apmltc.
  apply lt_0_Sn.
Qed.

(* Lemmas copied from gcd_p *)

Lemma gcd_comm : ∀ a b, gcd a b = gcd b a.
Proof.
intros a b.
apply gcd_p__f_gcd, gcd_p_comm, f_gcd__gcd_p.
reflexivity.
Qed.

Lemma gcd_mult : ∀ a b g, g = gcd a b → (∃k : nat, a = k * g).
Proof.
intros a b g H.
apply gcd_p_mult with (b := b).
apply f_gcd__gcd_p, H.
Qed.

Lemma gcd_0 : ∀ a, gcd a 0 = a.
Proof.
intros a.
unfold gcd.
reflexivity.
Qed.

Lemma gcd_is_zero : ∀ a b, gcd a b = 0 → a = 0 ∧ b = 0.
Proof.
intros a b H.
apply gcd_p_is_zero, f_gcd__gcd_p;
symmetry; assumption.
Qed.

Lemma mod_small : ∀ a b c, a < b → a mod b = c → c = a.
Proof.
intros a b c H₁ H₂.
eapply mod_p_by_gt; [ eassumption | idtac ].
apply f_mod__mod_p; symmetry; assumption.
Qed.

(* *)

Lemma gcd_le_a : ∀ a b, 0 < a → gcd a b ≤ a.
Proof.
intros a b a_pos.
remember (gcd a b) as g.
apply gcd_mult in Heqg.
destruct Heqg as (k, a_eq_kg).
subst a.
destruct k.
 apply lt_irrefl in a_pos; destruct a_pos.
 apply le_plus_l.
Qed.

(* gcd assoc *)

Definition divisible a b := f_mod b a = 0.

Lemma bezout_one_case : ∀ a b, 0 < a → ∃ u, ∃ v, a * u = b * v + gcd a b.
Proof.
intros a b Hapos.
remember (f_bezout a b) as bz.
destruct bz as (is_neg, uv).
destruct uv as (u', v').
symmetry in Heqbz; apply bezout_gcd in Heqbz.
destruct Heqbz as [H1| H1].
 destruct H1 as (His_neg, H1).
 exists u'; exists v'; exact H1.

 destruct H1 as (His_neg, H1).
 remember (gcd a b) as g.
 remember Heqg as Heqg'; clear HeqHeqg'.
 apply gcd_mult in Heqg.
 destruct Heqg as (a₁, H2).
 rewrite gcd_comm in Heqg'.
 apply gcd_mult in Heqg'.
 destruct Heqg' as (b₁, H3).
 exists ((a₁ - 1) * u' + 1).
 exists ((a₁ - 1) * v').
 subst a b.
 rewrite mult_plus_distr_l.
 rewrite mult_assoc.
 rewrite mult_assoc.
 rewrite mult_1_r.
 replace (b₁ * g * (a₁ - 1) * v') with (b₁ * g * v' * (a₁ - 1)).
  rewrite H1.
  rewrite mult_plus_distr_r.
  replace (a₁ * g * (a₁ - 1) * u') with (a₁ * g * u' * (a₁ - 1)).
   assert (g * (a₁ - 1) + g = a₁ * g) as H2.
    destruct a₁.
     apply lt_irrefl in Hapos; destruct Hapos.

     remember (S a₁ - 1) as x.
     assert (S a₁ = x + 1) as H2.
      rewrite Heqx.
      rewrite plus_comm.
      apply le_plus_minus.
      apply le_n_S, le_0_n.

      rewrite H2.
      rewrite mult_plus_distr_r, mult_1_l, mult_comm.
      reflexivity.

    rewrite <- plus_assoc, H2; reflexivity.

   rewrite <- mult_assoc.
   replace (u' * (a₁ - 1)) with ((a₁ - 1) * u').
    apply mult_assoc.
    apply mult_comm.

  rewrite <- mult_assoc.
  replace (v' * (a₁ - 1)) with ((a₁ - 1) * v').
   apply mult_assoc.
   apply mult_comm.
Qed.

Lemma div_a_b__div_gcd : ∀ a b k,
  divisible k a ∧ divisible k b → divisible k (gcd a b).
Proof.
unfold divisible.
intros a b k div_div.
destruct k.
 symmetry .
 apply mod_p__f_mod, m_a.
 
 assert (0 < S k) as k_pos by apply lt_0_Sn.
 remember (S k) as x; clear k Heqx; rename x into k.
 destruct div_div as (div_k_a, div_k_b).
 apply f_mod_is_zero in div_k_a; try assumption.
 apply f_mod_is_zero in div_k_b; try assumption.
 destruct div_k_a as (k₁, a_eq_k₁k).
 destruct div_k_b as (k₂, b_eq_k₂k).
 destruct a.
  remember (gcd 0 b) as g.
  rewrite gcd_comm in Heqg.
  rewrite gcd_0 in Heqg.
  subst.
  symmetry .
  apply mod_p__f_mod, eucl_div__mod.
  split; try assumption.
  exists k₂.
  rewrite plus_0_r.
  apply mult_comm.
  
  assert (0 < S a) as a_pos by apply lt_0_Sn.
  remember (S a) as a'.
  clear a Heqa'.
  rename a' into a.
  pose proof (bezout_one_case a b a_pos) as bez.
  destruct bez as (u, bez).
  destruct bez as (v, bez).
  rewrite mult_comm in a_eq_k₁k.
  rewrite mult_comm in b_eq_k₂k.
  subst a b.
  rewrite <- mult_assoc in bez.
  symmetry  in bez.
  rewrite <- mult_assoc in bez.
  symmetry  in bez.
  apply plus_minus in bez.
  rewrite <- mult_minus_distr_l in bez.
  rewrite bez.
  symmetry .
  apply mod_p__f_mod, eucl_div__mod.
  split; try assumption.
  exists (k₁ * u - k₂ * v).
  rewrite plus_0_r; reflexivity.
Qed.

Lemma mod_ka_a_0 : ∀ k a, f_mod (k * a) a = 0.
Proof.
intros k a.
destruct a.
 rewrite mult_0_r.
 reflexivity.
 
 symmetry .
 apply mod_p__f_mod, eucl_div__mod.
 split.
  exists k; rewrite plus_0_r; apply mult_comm.
  
  apply lt_0_Sn.
Qed.

Lemma gcd_assoc : ∀ a b c, gcd a (gcd b c) = gcd (gcd a b) c.
Proof.
intros a b c.
remember (gcd a (gcd b c)) as g₁.
remember (gcd (gcd a b) c) as g₂.
destruct g₁.
 apply f_gcd__gcd_p, gcd_p_is_zero in Heqg₁.
 destruct Heqg₁; subst a.
 remember (gcd 0 b) as g₁.
 rewrite gcd_comm, gcd_0 in Heqg₁.
 subst; symmetry ; assumption.
 
 assert (0 < S g₁) as g₁_pos by apply lt_0_Sn.
 remember (S g₁) as x; clear g₁ Heqx; rename x into g₁.
 destruct g₂.
  apply f_gcd__gcd_p, gcd_p_is_zero in Heqg₂.
  destruct Heqg₂; subst c.
  remember (gcd b 0) as g₂.
  rewrite gcd_0 in Heqg₂.
  subst; assumption.
  
  assert (0 < S g₂) as g₂_pos by apply lt_0_Sn.
  remember (S g₂) as x; clear g₂ Heqx; rename x into g₂.
  assert (divisible g₁ g₂) as div_g₁_g₂.
   rewrite Heqg₂.
   apply div_a_b__div_gcd; try assumption.
   split.
    apply div_a_b__div_gcd; try assumption.
    split.
     apply gcd_mult in Heqg₁.
     destruct Heqg₁ as (k, a_eq_kg₁).
     unfold divisible; subst a; apply mod_ka_a_0.
     
     rewrite gcd_comm in Heqg₁.
     apply gcd_mult in Heqg₁.
     destruct Heqg₁ as (k, g_eq_kg₁).
     symmetry  in g_eq_kg₁.
     apply gcd_mult in g_eq_kg₁.
     destruct g_eq_kg₁ as (k₁, b_eq_kk₁g₁).
     rewrite mult_assoc in b_eq_kk₁g₁.
     unfold divisible; subst b; apply mod_ka_a_0.
     
    rewrite gcd_comm in Heqg₁.
    apply gcd_mult in Heqg₁.
    destruct Heqg₁ as (k, g_eq_kg₁).
    symmetry  in g_eq_kg₁.
    rewrite gcd_comm in g_eq_kg₁.
    apply gcd_mult in g_eq_kg₁.
    destruct g_eq_kg₁ as (k₁, c_eq_kk₁g₁).
    rewrite mult_assoc in c_eq_kk₁g₁.
    unfold divisible; subst c; apply mod_ka_a_0.
    
   assert (divisible g₂ g₁) as div_g₂_g₁.
    rewrite Heqg₁.
    apply div_a_b__div_gcd; try assumption.
    split.
     apply gcd_mult in Heqg₂.
     destruct Heqg₂ as (k, g_eq_kg₂).
     symmetry  in g_eq_kg₂.
     apply gcd_mult in g_eq_kg₂.
     destruct g_eq_kg₂ as (k₁, a_eq_kk₁g₂).
     rewrite mult_assoc in a_eq_kk₁g₂.
     unfold divisible; subst a; apply mod_ka_a_0.
     
     apply div_a_b__div_gcd; try assumption.
     split.
      apply gcd_mult in Heqg₂.
      destruct Heqg₂ as (k, g_eq_kg₂).
      symmetry  in g_eq_kg₂.
      rewrite gcd_comm in g_eq_kg₂.
      apply gcd_mult in g_eq_kg₂.
      destruct g_eq_kg₂ as (k₁, b_eq_kk₁g₂).
      rewrite mult_assoc in b_eq_kk₁g₂.
      unfold divisible; subst b; apply mod_ka_a_0.
      
      rewrite gcd_comm in Heqg₂.
      apply gcd_mult in Heqg₂.
      destruct Heqg₂ as (k, c_eq_kg₂).
      unfold divisible; subst c; apply mod_ka_a_0.
      
    unfold divisible in div_g₁_g₂, div_g₂_g₁.
    apply f_mod_is_zero in div_g₁_g₂; try assumption.
    apply f_mod_is_zero in div_g₂_g₁; try assumption.
    destruct div_g₁_g₂ as (k₁, g₂_k₁g₁).
    destruct div_g₂_g₁ as (k₂, g₁_k₂g₂).
    assert (k₁ * g₁ - 1 * g₂ = 0) as H.
     rewrite g₂_k₁g₁,  mult_1_l; apply minus_diag.
     
     rewrite g₁_k₂g₂ in H.
     rewrite mult_assoc in H.
     rewrite <- mult_minus_distr_r in H.
     apply mult_is_O in H.
     destruct H as [k₂k₁_eq_1| g₁_eq_0].
      assert (1 + (k₁ * k₂ - 1) = k₁ * k₂) as H1.
       apply le_plus_minus_r.
       destruct k₁.
        rewrite g₂_k₁g₁ in g₂_pos; apply lt_irrefl in g₂_pos; destruct g₂_pos.
        
        destruct k₂.
         rewrite g₁_k₂g₂ in g₁_pos; apply lt_irrefl in g₁_pos;
          destruct g₁_pos.
         
         simpl.
         apply le_n_S, le_0_n.
         
       rewrite k₂k₁_eq_1 in H1.
       symmetry  in H1; rewrite plus_0_r in H1.
       apply mult_is_one in H1.
       destruct H1; subst k₂.
       rewrite mult_1_l in g₁_k₂g₂; assumption.
       
      rewrite g₁_eq_0 in g₂_pos; apply lt_irrefl in g₂_pos; destruct g₂_pos.
Qed.

(* *)

Lemma mod_mod_plus : ∀ a b c,
  f_mod (a + b) c = f_mod (f_mod a c + b) c.
Proof.
intros a b c.
induction b.
 rewrite plus_0_r.
 rewrite plus_0_r.
 remember (f_mod a c) as m.
 apply mod_p__f_mod, mod_idemp with (a := a), f_mod__mod_p.
 assumption.

 simpl.
 rewrite <- plus_n_Sm.
 rewrite <- plus_n_Sm.
 simpl.
 remember (f_mod (a + b) c) as m.
 remember (f_mod (f_mod a c + b) c) as n.
 destruct (lt_dec (S m) c).
  destruct (lt_dec (S n) c).
   rewrite IHb; reflexivity.
   rewrite IHb in l; contradiction .
  destruct (lt_dec (S n) c).
   rewrite IHb in n0; contradiction .
   reflexivity.
Qed.

Lemma mod_plus_lt : ∀ a b c,
  f_mod a c + f_mod b c < c → f_mod (a + b) c = f_mod a c + f_mod b c.
Proof.
intros a b c H1.
rewrite mod_mod_plus.
symmetry ; apply plus_mod_lt.
assumption.
Qed.
