(* $Id: gcd_p.v,v 1.8 2011-12-01 10:03:59 deraugla Exp $ *)

Require Import Utf8.
Require Import Arith.
Require Import Misc.

Inductive mod_p : nat → nat → nat → Prop :=
  | m_a : ∀ a, mod_p a 0 0
  | m_b : ∀ b, mod_p 0 b 0
  | m_r : ∀ a b r, S r < b → mod_p a b r → mod_p (S a) b (S r)
  | m_q : ∀ a r, mod_p a (S r) r → mod_p (S a) (S r) 0.

Inductive gcd_p : nat → nat → nat → Prop :=
  | gcd_O : ∀ a, gcd_p a 0 a
  | gcd_Sn : ∀ a b r, ∀ g, 0 < b → mod_p a b r → gcd_p b r g → gcd_p a b g.

(* mod *)

Theorem mod_p_by_gt : ∀ a b c, a < b → mod_p a b c → c = a.
Proof.
intros a b c H1 H2.
induction H2.
 inversion H1.
  
 reflexivity.
 
 apply eq_S, IHmod_p.
 apply lt_le_weak, H1.
 
 apply lt_S_n in H1.
 remember H1 as H3; clear HeqH3.
 apply lt_S, IHmod_p in H1.
 rewrite H1 in H3.
 apply lt_irrefl in H3.
 inversion H3.
Qed.

Theorem lt_mod : ∀ a b, a < b → mod_p a b a.
Proof.
  intros a b H1.
  induction a.
   apply m_b.
   apply m_r.
    apply H1.
    apply IHa, le_S_n, le_S, H1.
Qed.

Theorem mod_by_0 : ∀ a b, mod_p a 0 b → b = 0.
Proof.
intros a b H1.
inversion H1 as [| | a1 b1 r1 H2 H3| ]; subst.
 split; reflexivity.
 reflexivity.
 inversion H2.
Qed.

Theorem mod_lt : ∀ a b c, mod_p a b c → c < b \/ c = 0.
Proof.
intros a b c H1.
inversion H1 as [| |a1 b1 r1 H2 H3 H4 H5 H6| ]; subst.
 right; reflexivity.
 right; reflexivity.
 left; exact H2.
 left; apply lt_0_Sn.
Qed.

Theorem mod_le_a : ∀ a b c, mod_p a b c → c ≤ a.
Proof.
intros a b c H1.
induction H1.
 apply le_0_n.
 apply le_n.
 apply le_n_S, IHmod_p.
 apply le_0_n.
Qed.

Theorem mod_idemp : ∀ a b r, mod_p a b r → mod_p r b r.
Proof.
intros a b r H1.
apply mod_lt in H1.
destruct H1 as [H1| H1].
 apply lt_mod, H1.
 subst; apply m_b.
Qed.

Theorem mod__eucl_div : ∀ a b r,
  0 < b → mod_p a b r → (∃ q, a = b * q + r) ∧ r < b.
Proof.
intros a b r Hb H1; split.
 induction H1.
  inversion Hb.

  exists 0; rewrite mult_0_r; reflexivity.

  apply IHmod_p in Hb.
  destruct Hb as (x, Hb).
  exists x; subst a.
  rewrite plus_n_Sm; reflexivity.

  apply IHmod_p in Hb.
  destruct Hb as (x, Hb).
  exists (x + 1); subst a.
  rewrite plus_0_r, mult_plus_distr_l, plus_n_Sm, mult_1_r; reflexivity.

 inversion H1 as [a2 H2 | b2 H2 | a2 b2 r2 H2 H3 | a2 r2 H2]; subst.
  exact Hb.
  exact Hb.
  exact H2.
  exact Hb.
Qed.

Theorem eucl_div__mod : ∀ a b r, (∃ q, a = b * q + r) ∧ r < b → mod_p a b r.
Proof.
induction a; intros b r (H, H2); inversion_clear H as (q, H1).
 symmetry in H1.
 apply plus_is_O in H1.
 destruct H1 as (H3, H4).
 rewrite H4; apply m_b.

 destruct r.
  destruct q.
   rewrite mult_0_r, plus_0_r in H1; inversion H1.

   destruct b.
    inversion H2.

    apply m_q, IHa; split.
     exists q.
     rewrite plus_0_r in H1.
     rewrite mult_comm in H1.
     simpl in H1.
     apply eq_add_S in H1.
     rewrite plus_comm, mult_comm.
     exact H1.

     apply le_n.

  apply m_r, IHa.
   apply H2.
   split.
    exists q.
    rewrite <- plus_n_Sm in H1.
    inversion H1 as (H3).
    reflexivity.

    apply le_S_n, le_S, H2.
Qed.

Theorem mod_plus_b__mod : ∀ a b r, mod_p (a+b) b r → mod_p a b r.
Proof.
intros a b r H1.
destruct b.
 rewrite plus_0_r in H1; exact H1.

 apply mod__eucl_div in H1.
  destruct H1 as (H1, H2).
  destruct H1 as (q, H1).
  destruct q.
   rewrite mult_0_r in H1; simpl in H1.
   rewrite <- plus_0_r in H2.
   subst r.
   rewrite plus_comm in H2.
   apply plus_lt_reg_l in H2.
   apply lt_n_O in H2; destruct H2.

   rewrite mult_comm in H1.
   simpl in H1.
   rewrite <- plus_assoc in H1.
   rewrite <- plus_Sn_m in H1.
   rewrite plus_comm in H1.
   apply plus_reg_l in H1.
   rewrite mult_comm in H1.
   apply eucl_div__mod.
   split; [ exists q; apply H1 | apply H2 ].

  apply lt_0_Sn.
Qed.

Theorem mod_plus_kb__mod : ∀ a b r k, mod_p (a+k*b) b r → mod_p a b r.
Proof.
intros a b r k H1.
induction k.
 rewrite plus_0_r in H1; exact H1.

 apply IHk, mod_plus_b__mod.
 rewrite <- plus_assoc.
 rewrite mult_comm.
 rewrite mult_n_Sm.
 rewrite mult_comm.
 exact H1.
Qed.

Lemma lt_mult : ∀ a b c, a < b → a = b * c → a = 0.
Proof.
  intros a b c H1 H2.
  destruct c.
   rewrite <- mult_n_O in H2; apply H2.

   rewrite H2 in H1.
   rewrite <- mult_n_Sm, <- plus_0_r in H1.
   rewrite plus_comm in H1.
   apply plus_lt_reg_l in H1; inversion H1.
Qed.

Lemma mod_uniq_le : ∀ a b c d, c ≤ d → mod_p a b c → mod_p a b d → c = d.
Proof.
intros a b c d H4 H1 H2.
destruct b.
 apply mod_by_0 in H1.
 apply mod_by_0 in H2.
 destruct H1, H2; subst; reflexivity.

 rename b into b'; remember (S b') as b.
 remember H1 as H8.
 remember H2 as H9.
 clear HeqH8 HeqH9.
 apply mod_lt in H8.
 apply mod_lt in H9.
 apply mod__eucl_div in H1.
  inversion_clear H1 as (H'1, H''1).
  inversion H'1; subst.
  apply mod__eucl_div in H2.
   inversion_clear H2 as (H'2, H''2).
   inversion H'2.
   remember H4 as H7.
   clear HeqH7.
   apply le_plus_minus in H4.
   rewrite H4 in H.
   rewrite plus_comm in H.
   symmetry  in H.
   rewrite plus_comm in H.
   rewrite <- plus_assoc in H.
   apply plus_reg_l in H.
   rewrite plus_comm in H.
   symmetry  in H.
   apply plus_minus in H.
   rewrite <- mult_minus_distr_l in H.
   apply lt_mult in H.
    apply minus_le in H.
    apply le_antisym in H; [ apply H | apply H7 ].

    apply lt_minus_weak; exact H''2.

   apply lt_0_Sn.

  subst b; apply lt_0_Sn.
Qed.

Theorem mod_uniq : ∀ a b c d, mod_p a b c → mod_p a b d → c = d.
Proof.
intros a b c d H1 H2.
pose proof (le_lt_dec c d) as H3.
inversion_clear H3 as [H4| H4].
 apply mod_uniq_le with (a := a) (b := b); [apply H4 | apply H1 | apply H2].
 apply lt_le_weak in H4.
 symmetry.
 apply mod_uniq_le with (a := a) (b := b); [apply H4 | apply H2 | apply H1].
Qed.

(* gcd *)

Theorem gcd_p_comm : ∀ a b g, gcd_p a b g → gcd_p b a g.
Proof.
intros a b g H1.
pose proof (lt_eq_lt_dec a b) as H2.
inversion_clear H2 as [H3| H3].
 inversion_clear H3 as [H2| H2].
  inversion_clear H1 as [H3 H4| a1 b1 r g1 H3 H4].
   destruct g.
    apply gcd_O.
    apply gcd_Sn with (r := 0).
     apply lt_0_Sn.
     apply m_b.
     apply gcd_O.
   apply mod_p_by_gt in H4.
    subst; exact H.
    exact H2.
  subst; exact H1.
 apply gcd_Sn with (r := b).
  clear g H1.
  destruct b.
   exact H3.
   destruct a.
    inversion H3.
    apply lt_0_Sn.
  apply eucl_div__mod.
  split.
   exists 0; rewrite mult_0_r; reflexivity.
   apply H3.
  exact H1.
Qed.

Theorem gcd_uniq : ∀ a b g1 g2, gcd_p a b g1 → gcd_p a b g2 → g1 = g2.
Proof.
intros a b g1 g2 H1 H2.
induction H1 as [| a b r g H3 H4 H5].
 inversion H2 as [| a1 b r g H3 H4 H5]; subst.
  reflexivity.
  inversion H3.
 inversion H2 as [| a1 b1 r1 g1 H6 H7 H8]; subst.
  inversion H3.
  apply mod_uniq with (c := r1) in H4; subst; [ apply IHH5, H8 | apply H7 ].
Qed.

(* *)

Inductive gcdc_p : nat → nat → nat → nat → Prop :=
  | gcdc_0 : ∀ a, gcdc_p a 0 a 0
  | gcdc_n : ∀ a b r, ∀ g n,
      0 < b → mod_p a b r → gcdc_p b r g n → gcdc_p a b g (S n).

Theorem nat_ind2 : ∀ P : nat → Prop,
  P 0 → P 1 → (∀ n, P n → P (S n) → P (S (S n))) → ∀ n, P n.
Proof.
  intros P P0 P1 Pn.
  fix F 1.
  intros n.
  destruct n.
   apply P0.
   induction n.
    apply P1.
    apply Pn; [apply F | apply IHn].
Qed.

Theorem gcd_p__gcdc_p : ∀ a b g, gcd_p a b g → ∃ n, gcdc_p a b g n.
Proof.
intros a b g H1.
induction H1.
 exists 0; apply gcdc_0.
 destruct IHgcd_p as (m, H2).
 apply gcdc_n with (a := a) in H2.
  exists (S m); apply H2.
  apply H.
  exact H0.
Qed.

Theorem gcdc_mult : ∀ r₁ r₂ g n, gcdc_p r₁ r₂ g n → ∃ k₁, r₁ = k₁ * g.
Proof.
intros r₁ r₂ g n; revert r₁ r₂ g.
induction n using nat_ind2; intros r₁ r₂ g H.
 inversion H as [r₃| x₁ x₂ r₃ x₃ n H1 H2 H3]; subst.
 exists 1; rewrite mult_1_l; reflexivity.

 inversion_clear H as [x₁| x₁ x₂ r₃ x₃ x₄ H1 H2 H3].
 inversion H3; subst.
 apply mod__eucl_div in H2.
  destruct H2 as (H4, H5).
  destruct H4 as (q₂, H₁).
  exists q₂; rewrite H₁, plus_0_r; apply mult_comm.

  exact H1.

 inversion_clear H as [x₁| x₁ x₂ r₃ x₃ x₄ H1 H2 H3].
 inversion H3 as [x₁| x₁ x₂ r₄ x₄ n₁ H4 H5 H6 H7]; subst.
 apply mod__eucl_div in H2.
  destruct H2 as (H2, LT₃).
  destruct H2 as (q₂, H₁).
  apply IHn0 in H3.
  destruct H3 as (q₃, H₂).
  apply IHn in H8.
  destruct H8 as (q₄, H₃).
  rewrite H₂, H₃ in H₁.
  exists (q₂ * q₃ + q₄).
  rewrite mult_plus_distr_r, <- mult_assoc, mult_comm.
  exact H₁.

  exact H1.
Qed.

Theorem gcd_p_mult : ∀ a b g, gcd_p a b g → ∃ k, a = k * g.
Proof.
intros a b g H1.
apply gcd_p__gcdc_p in H1.
destruct H1 as (n, H1).
apply gcdc_mult in H1.
destruct H1 as (k, H1).
exists k; apply H1.
Qed.

Theorem gcd_p_is_zero : ∀ a b, gcd_p a b 0 → a = 0 ∧ b = 0.
Proof.
intros a b H1.
split.
 apply gcd_p_mult in H1.
 destruct H1 as (k, H1).
 rewrite mult_0_r in H1; exact H1.

 apply gcd_p_comm in H1.
 apply gcd_p_mult in H1.
 destruct H1 as (k, H1).
 rewrite mult_0_r in H1; exact H1.
Qed.
